"""
Python script by Tobias Polzin, polzin@gmx.de
All rights reserved.

<title>
scannew.py, Version 22.12.99
</title>
<long>
Module for symbolic calculation with python (uses PYTHONICA). Example:
<pre>
>>> print sym("x")/2.0+10*sym("x")+sym("x")/sym("x")
10.5 x + 1
</pre>
</long>
"""
from pythonicacore import *
from types import *

gSymToName = {}
gNameToSym = {}
gNameToPrec = {}


#----------------------------------------------------------------------
def buildOperTables():
	global gSymToName, gNameToPrec
	ops = ( '/.',	'ReplaceAll',
			'->',	'Rule',
			'==',	'Equal',
			'=',	'Set',
			'=.',	'Unset',
			'+',	'Plus',
			'-',	'Minus',
			'*',	'Times',
			'/',	'Divide',
			'^',	'Power')
	for i in range(0,len(ops)/2):
		gSymToName[ops[i*2]] = ops[i*2+1]
		gNameToSym[ops[i*2+1]] = ops[i*2]
		gNameToPrec[ops[i*2+1]] = i
buildOperTables()

#----------------------------------------------------------------------
def unparse(expr, leftprec=-1, rightprec=-1):
	# is this expression something we have an operator for?
	head = expr.Head()
	if head not in gNameToPrec.keys():
		# nope, no operator... just print normally
		out = str(expr)
		return out
	# yep, we have an operator... get it, and its precedence
	prec = gNameToPrec[head]
	# (for readability, we'll code a couple of formatting hacks)
	op = gNameToSym[head]
	if op == '*': op = ' '
	elif op != '^': op = ' ' + op + ' '
	# combine data using operator -- pass precedence info
	out = ''
	# leftmost argument gets has our left, and us for the right
	if len(expr.data) > 0:
		if len(expr.data) == 1: r = rightprec
		else: r = prec
		out = unparse(expr.data[0],leftprec,r)
	# middle arguments have us on both left and right
	if len(expr.data) > 2:
		out = out + op + \
		  string.join(map(lambda x,p=prec:unparse(x,p,p),expr.data[1:-1]), \
		 		 op)
	# last argument has our right, and us for the left
	if len(expr.data) > 1:
		out = out + op + \
				unparse(expr.data[-1],prec,rightprec)
	# use parenthesis if we're lower precedence than our neighbors
	if prec < leftprec or prec < rightprec:
		out = '(' + out + ')'
	return out

class sym:
    def __init__(self,a):
        t=type(a)
        if t==StringType:
            a=ExprSymbol(a)
        elif t in [IntType,FloatType,LongType]:
            a=ExprNumber(a)
        elif a.__class__ == sym:
            a=a.s
        if type(a)!=InstanceType:
            print "ERROR!"
        self.s=a
    def tiny(self,debug=0):
        if debug:
            print self.s," --> ",
        self.s=self.s.Eval()
        while 1:
            simp=self.s.Simplify()
            if simp==self.s: break
            self.s=simp
        if debug:
            print self.s
        return self
    def __str__(self):
        return unparse(self.tiny().s)
    def __repr__(self):
        return "sym("+`self.tiny().s`+")"
    def __add__(self, o): return sym(ExprPlus([self.s,sym(o).s]))
    def __mul__(self, o): return sym(ExprTimes([self.s,sym(o).s]))
    __rmul__ = __mul__
    __radd__ = __add__
    def __sub__(self, o): return sym(ExprMinus([self.s,sym(o).s]))
    def __div__(self, o): return sym(ExprDivide([self.s,sym(o).s]))

#b=sym("x")/2.0+10*sym("x")+sym("x")/sym("x")
#print "str:",b
#  str: 10.5 x + 1


